/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* Copyright (c) 2016 King Kévin <kingkevin@cuvoodoo.info> */
/* this library handles IrDA SIR (USART based) communication */
/* this library handles IrDA communication */

/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdio.h> // standard I/O facilities
#include <stdlib.h> // general utilities

/* STM32 (including CM3) libraries */
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/usart.h> // universal synchronous asynchronous receiver transmitter library
#include <libopencm3/cm3/nvic.h> // interrupt handler
#include <libopencmsis/core_cm3.h> // Cortex M3 utilities

#include "usart_irda.h" // IrDA header and definitions

/* which USART to use for IrDA */
#define IRDA USART3
#define IRDA_RCC RCC_USART3
#define IRDA_IRQ NVIC_USART3_IRQ
#define IRDA_PORT GPIOB
#define IRDA_PIN_TX GPIO_USART3_TX
#define IRDA_PIN_RX GPIO_USART3_RX
/* serial baudrate, in bits per second (with 8N1 8 bits, no parity bit, 1 stop bit settings) */
#define IRDA_BAUDRATE 9600
/* RX and TX buffer sizes */
#define IRDA_BUFFER 128

/* input and output ring buffer, indexes, and available memory */
static uint8_t rx_buffer[IRDA_BUFFER] = {0};
static volatile uint8_t rx_i = 0;
static volatile uint8_t rx_used = 0;
static uint8_t tx_buffer[IRDA_BUFFER] = {0};
static volatile uint8_t tx_i = 0;
static volatile uint8_t tx_used = 0;
/* show the user how much data received over IrDA is ready */
volatile uint8_t irda_received = 0; // same as rx_used, but since the user can write this variable we don't rely on it

/* setup IrDA peripheral */
void irda_setup(void)
{
	rcc_periph_clock_enable(IRDA_RCC); // enable clock for USART/IrDA block
	gpio_set_mode(IRDA_PORT, GPIO_MODE_OUTPUT_10_MHZ, GPIO_CNF_OUTPUT_ALTFN_PUSHPULL, IRDA_PIN_TX); // setup GPIO pin USART/IrDA transmit
	gpio_set_mode(IRDA_PORT, GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, IRDA_PIN_RX); // setup GPIO pin USART/IrDA receive
	gpio_set(IRDA_PORT, IRDA_PIN_RX); // pull up to avoid noise when not connected

	/* setup UART/IrDA parameters */
	usart_set_baudrate(IRDA, IRDA_BAUDRATE);
	usart_set_databits(IRDA, 8);
	usart_set_stopbits(IRDA, USART_STOPBITS_1);
	usart_set_mode(IRDA, USART_MODE_TX_RX);
	usart_set_parity(IRDA, USART_PARITY_NONE);
	usart_set_flow_control(IRDA, USART_FLOWCONTROL_NONE);
	USART_CR3(IRDA) |= USART_CR3_IREN; // enable IrDA SIR ENDEC block (using IREN)

	nvic_enable_irq(IRDA_IRQ); // enable the USART/IrDA interrupt
	usart_enable_rx_interrupt(IRDA); // enable receive interrupt
	usart_enable(IRDA); // enable USART/IrDA

	/* reset buffer states */
	tx_i = 0;
	tx_used = 0;
	rx_i = 0;
	rx_used = 0;
	irda_received = 0;
}

/* put character on IrDA (blocking) */
void irda_putchar_blocking(char c)
{
	irda_flush(); // empty buffer first
	usart_send_blocking(IRDA, c); // send character
}

/* ensure all data has been transmitted (blocking) */
void irda_flush(void)
{
	while (tx_used) { // idle until buffer is empty
		__WFI(); // sleep until interrupt
	}
	usart_wait_send_ready(IRDA); // wait until transmit register is empty (transmission might not be complete)
}

/* get character from IrDA (blocking) */
char irda_getchar(void)
{
	while (!rx_used) { // idle until data is available
		__WFI(); // sleep until interrupt;
	}
	char to_return = rx_buffer[rx_i]; // get the next available character
	rx_i = (rx_i+1)%sizeof(rx_buffer); // update used buffer
	rx_used--; // update used buffer
	irda_received = rx_used; // update available data
	return to_return;
}

/* put character on IrDA (non-blocking until buffer is full) */
void irda_putchar_nonblocking(char c)
{
	while (tx_used>=sizeof(tx_buffer)) { // idle until buffer has some space
		usart_enable_tx_interrupt(IRDA); // enable transmit interrupt
		__WFI(); // sleep until something happened
	}
	tx_buffer[(tx_i+tx_used)%sizeof(tx_buffer)] = c; // put character in buffer
	tx_used++; // update used buffer
	usart_enable_tx_interrupt(IRDA); // enable transmit interrupt
}

#if (IRDA==USART1)
void usart1_isr(void)
#elif (IRDA==USART2)
void usart2_isr(void)
#elif (IRDA==USART3)
void usart3_isr(void)
#endif
{ // USART interrupt
	if (usart_get_interrupt_source(IRDA, USART_SR_TXE)) { // data has been transmitted
		if (!tx_used) { // no data in the buffer to transmit
			usart_disable_tx_interrupt(IRDA); // disable transmit interrupt
		} else {
			usart_send(IRDA,tx_buffer[tx_i]); // put data in transmit register
			tx_i = (tx_i+1)%sizeof(rx_buffer); // update location on buffer
			tx_used--; // update used size
		}
	}
	if (usart_get_interrupt_source(IRDA, USART_SR_RXNE)) { // data has been received
		// only save data if there is space in the buffer
		if (rx_used>=sizeof(rx_buffer)) {
			usart_recv(IRDA); // read to clear interrupt
		} else {
			rx_buffer[(rx_i+rx_used)%sizeof(rx_buffer)] = usart_recv(IRDA); // put character in buffer
			rx_used++; // update used buffer
			irda_received = rx_used; // update available data
		}
	}
}
